//The code is at PoC level, it was developed for demonstration purposes.
//The code is released under GPL (http://www.gnu.org/licenses/gpl.html), by Laszlo Toth.
//Use the code at your own responsibility.
//http://soonerorlater.hu/index.khtml?article_id=516

#include <string.h>
#include <openssl/evp.h> 
#include <openssl/sha.h>
#include <openssl/md5.h>
#include "getopt.h"

int HexStringtoBinArray(unsigned char* str, unsigned char* array);
int decrypt_block(unsigned char* block, unsigned char* key, unsigned char* iv, int block_size, EVP_CIPHER* cipher, unsigned char* decrypted_block);

static struct option long_options[] ={
	{"file", required_argument, 0, 'f'},
	{"key", required_argument, 0, 'k'},
	{"ofile", required_argument, 0, 'o'},
	{0,0,0,0}
};

int main(int argc, char* argv[])
{
	char* ifile=NULL;
	FILE* ifh;
	char* ofile=NULL;
	FILE* ofh;
	char* testkey=NULL;
	unsigned char bintestencdata[10000];
	unsigned char bintestkey[100];
	unsigned char bintestiv[100];
	int block_size=8192;
	unsigned char* block;
	unsigned char* decrypted_block;
	EVP_CIPHER* cipher;

	int option_index = 0;
	int opt = getopt_long (argc, argv, "f:k:o:",
		long_options, &option_index);

	while(opt != -1){
		switch(opt){
		case 'f':
			ifile=optarg;
			break;
		case 'k':
			testkey=optarg;
			break;
		case 'o':
			ofile=optarg;
			break;
		}
		opt = getopt_long (argc, argv, "f:k:o:", long_options, &option_index);
	}

	if(!testkey){
		printf("Key parameter (-k) is missing!\n");
		return -1;
	}

	int keylen=strlen(testkey);
	switch(keylen){
	case 32:
		cipher=(EVP_CIPHER*)EVP_aes_128_cfb128();
		break;
	case 48:
		cipher=(EVP_CIPHER*)EVP_aes_192_cfb128();
		break;
	case 64:
		cipher=(EVP_CIPHER*)EVP_aes_256_cfb128();
		break;
	default:
		printf("Invalid key length!\nThe valid length are: 128, 168 (not supported yet), 192, 256 bit!\n");
		return -1;
	}

	HexStringtoBinArray((unsigned char*)testkey,bintestkey);

	if(!ifile){
		printf("Input file (-f) parameter is missing!\n");
		return -1;
	}
	ifh=fopen(ifile, "rb");
	if(!ifh){
		printf("Could not open the input file: %s!\n", ifile);
		return -1;
	}

	if(ofile){
		ofh=fopen(ofile, "wb");
		if(!ifh){
			printf("Could not open the output file: %s!\n", ofile);
			return -1;
		}
	}else{
		printf("Output file parameter (-o) is missing!\n");
		return -1;
	}

	//read the size of the blocks
	fseek(ifh, 20, SEEK_SET);
	fread(&block_size, 4, 1, ifh);

	fseek(ifh,0, SEEK_SET);

	block=(unsigned char*)calloc(block_size, sizeof(char));
	decrypted_block=(unsigned char*)calloc(block_size, sizeof(char));

	while(!feof(ifh)){
		int count=fread(block,1,block_size,ifh);
		if(count < 8192){
			break;
		}

		//block type
		unsigned char block_type;
		memmove(&block_type,block, 1);
		//what is in the block
		unsigned char data_type;
		memmove(&data_type, block+20,1);

		if(block_type==6){
			printf("Block type: %X, data type: %X\n", block_type, data_type);
			memset(bintestiv,0,16);
			memmove(bintestiv, block+4,8);
			memmove(bintestiv+12,block+14,1);
			decrypt_block(block+20, bintestkey, bintestiv, block_size, cipher, decrypted_block);
			fwrite(block,1,20,ofh);
			fwrite(decrypted_block,1, block_size-24, ofh);
			fwrite(block+block_size-4,1,4,ofh);
		}else{
			fwrite(block, 1, block_size,ofh);
		}
	}

	fclose(ifh);
	fclose(ofh);

	return 0;
}

int HexStringtoBinArray(unsigned char* str, unsigned char* array){

	int alen=strlen((char*)str)/2;
	unsigned char t[2];
	unsigned int hexc;
	int j=0;

	for(unsigned int i=0;i<strlen((char*)str);i=i+2){
		t[0]=str[i];
		t[1]=str[i+1];
		hexc = t[0]-48;
		if (hexc > 9) hexc-=7;
		array[j]=hexc*16;
		hexc = t[1]-48;
		if (hexc > 9) hexc-=7;
		array[j]+=hexc;
		j++;
	}
	return j;
}

int decrypt_block(unsigned char* block, unsigned char* key, unsigned char* iv, int block_size, EVP_CIPHER* cipher, unsigned char* decrypted_block){

	EVP_CIPHER_CTX ctx;
	int outlen=0;

	EVP_CIPHER_CTX_init(&ctx);
	EVP_DecryptInit_ex(&ctx, cipher, NULL, key, iv);
	EVP_DecryptUpdate(&ctx, decrypted_block, &outlen, block, block_size-24);
	EVP_CIPHER_CTX_cleanup(&ctx);
	return 0;
}